unit Generics.Trees;

interface

uses
  Generics.Collections;

type
  /// Rfrence  une routine de call-back avec un paramtre
  TValueCallBack<T> = reference to procedure(const Value: T);

  {*
    Structure arborescente gnrique
  *}
  TTreeNode<T> = class(TObject)
  private
    FParent: TTreeNode<T>;                /// Noeud parent (nil si racine)
    FChildren: TObjectList<TTreeNode<T>>; /// Liste des enfants
    FValue: T;                            /// Valeur tiquette

    FDestroying: Boolean; /// Indique si est en train d'tre dtruit

    procedure DoAncestorChanged;

    function GetRootNode: TTreeNode<T>;
    function GetChildCount: Integer; inline;
    function GetChildren(Index: Integer): TTreeNode<T>; inline;
    function GetIndexAsChild: Integer; inline;

    function GetIsRoot: Boolean; inline;
    function GetIsLeaf: Boolean; inline;
    function GetDepth: Integer;
  protected
    procedure AncestorChanged; virtual;
    procedure Destroying; virtual;

    procedure AddChild(Index: Integer; Child: TTreeNode<T>); virtual;
    procedure RemoveChild(Child: TTreeNode<T>); virtual;

    procedure SetValue(const AValue: T); virtual;

    property IsDestroying: Boolean read FDestroying;
  public
    constructor Create(AParent: TTreeNode<T>; const AValue: T); overload;
    constructor Create(AParent: TTreeNode<T>); overload;
    constructor Create(const AValue: T); overload;
    constructor Create; overload;
    destructor Destroy; override;

    procedure AfterConstruction; override;
    procedure BeforeDestruction; override;

    procedure MoveTo(NewParent: TTreeNode<T>; Index: Integer = -1); overload;
    procedure MoveTo(Index: Integer); overload;

    function IndexOf(Child: TTreeNode<T>): Integer; inline;

    procedure PreOrderWalk(
      const Action: TValueCallBack<TTreeNode<T>>); overload;
    procedure PreOrderWalk(const Action: TValueCallBack<T>); overload;

    procedure PostOrderWalk(
      const Action: TValueCallBack<TTreeNode<T>>); overload;
    procedure PostOrderWalk(const Action: TValueCallBack<T>); overload;

    property Parent: TTreeNode<T> read FParent;
    property RootNode: TTreeNode<T> read GetRootNode;
    property ChildCount: Integer read GetChildCount;
    property Children[Index: Integer]: TTreeNode<T> read GetChildren;
    property IndexAsChild: Integer read GetIndexAsChild;

    property IsRoot: Boolean read GetIsRoot;
    property IsLeaf: Boolean read GetIsLeaf;
    property Depth: Integer read GetDepth;

    property Value: T read FValue write SetValue;
  end;

  {*
    Structure arborescente gnrique dont les valeurs sont des objets
    Lorsque le noeud est libr, la valeur tiquete est libre galement.
  *}
  TObjectTreeNode<T: class> = class(TTreeNode<T>)
  public
    destructor Destroy; override;
  end;

  {*
    Structure arborescente gnrique dont les valeurs sont des objets
    Lorsqu'un noeud est cr sans valeur tiquete, une nouvelle valeur est
    cre avec le constructeur sans paramtre du type choisi.
    Lorsque le noeud est libr, la valeur tiquete est libre galement.
  *}
  TCreateObjectTreeNode<T: class, constructor> = class(TObjectTreeNode<T>)
  public
    constructor Create(AParent: TTreeNode<T>); overload;
    constructor Create; overload;
  end;

implementation

{--------------}
{ TTreeNode<T> }
{--------------}

{*
  Cre un noeud avec un parent et une valeur tiquete
  @param AParent   Parent
  @param AValue    Valeur tiquete
*}
constructor TTreeNode<T>.Create(AParent: TTreeNode<T>; const AValue: T);
begin
  inherited Create;

  FParent := AParent;
  FChildren := TObjectList<TTreeNode<T>>.Create;
  FValue := AValue;
end;

{*
  Cre un noeud avec un parent mais sans valeur tiquete
  @param AParent   Parent
*}
constructor TTreeNode<T>.Create(AParent: TTreeNode<T>);
begin
  Create(AParent, Default(T));
end;

{*
  Cre un noeud avec une valeur tiquete mais sans parent
  @param AValue   Valeur tiquete
*}
constructor TTreeNode<T>.Create(const AValue: T);
begin
  Create(nil, AValue);
end;

{*
  Cre un noeud sans parent ni valeur tiquete
*}
constructor TTreeNode<T>.Create;
begin
  Create(nil, Default(T));
end;

{*
  [@inheritDoc]
*}
destructor TTreeNode<T>.Destroy;
begin
  FChildren.Free;

  inherited;
end;

{*
  Appelle la procdure AncestorChanged sur toute la descendance (prfixe)
*}
procedure TTreeNode<T>.DoAncestorChanged;
begin
  PreOrderWalk(
    procedure(const Node: TTreeNode<T>)
    begin
      Node.AncestorChanged;
    end);
end;

{*
  Noeud racine de l'arbre
  @return Noeud racine de l'arbre
*}
function TTreeNode<T>.GetRootNode: TTreeNode<T>;
begin
  if IsRoot then
    Result := Self
  else
    Result := Parent.RootNode;
end;

{*
  Nombre de noeuds enfants
  @return Nombre de noeuds enfants
*}
function TTreeNode<T>.GetChildCount: Integer;
begin
  Result := FChildren.Count;
end;

{*
  Tableau zero-based des enfants
  @param Index   Index d'un enfant
  @return Enfant  l'index spcifi
*}
function TTreeNode<T>.GetChildren(Index: Integer): TTreeNode<T>;
begin
  Result := FChildren[Index];
end;

{*
  Index de ce noeud parmi les enfants de son parent
  @return Index de ce noeud parmi les enfants de son parent (-1 si la racine)
*}
function TTreeNode<T>.GetIndexAsChild: Integer;
begin
  if Parent = nil then
    Result := -1
  else
    Result := Parent.IndexOf(Self);
end;

{*
  Teste si ce noeud est la racine de l'arbre
  @return True si c'est la racine, False sinon
*}
function TTreeNode<T>.GetIsRoot: Boolean;
begin
  Result := Parent = nil;
end;

{*
  Teste si ce noeud est une feuille de l'arbre
  @return True si c'est une feuille, False sinon
*}
function TTreeNode<T>.GetIsLeaf: Boolean;
begin
  Result := FChildren.Count = 0;
end;

{*
  Profondeur du noeud
  @return Profondeur du noeud (la racine a une profondeur de 0)
*}
function TTreeNode<T>.GetDepth: Integer;
var
  Ancestor: TTreeNode<T>;
begin
  Ancestor := Parent;
  Result := 0;

  while Ancestor <> nil do
  begin
    Inc(Result);
    Ancestor := Ancestor.Parent;
  end;
end;

{*
  Mthode de notification appele lorsqu'un anctre a t modifi
*}
procedure TTreeNode<T>.AncestorChanged;
begin
end;

{*
  Mthode de notification appele lorsqu'un anctre va tre dtruit
*}
procedure TTreeNode<T>.Destroying;
begin
  FDestroying := True;
end;

{*
  Ajoute un enfant
  @param Child   Enfant  ajouter
*}
procedure TTreeNode<T>.AddChild(Index: Integer; Child: TTreeNode<T>);
begin
  if Index < 0 then
    FChildren.Add(Child)
  else
    FChildren.Insert(Index, Child);
end;

{*
  Retire un enfant
  @param Child   Enfant  retirer
*}
procedure TTreeNode<T>.RemoveChild(Child: TTreeNode<T>);
begin
  FChildren.Remove(Child);
end;

{*
  Modifie la valeur tiquete  ce noeud
  @param AValue   Nouvelle valeur
*}
procedure TTreeNode<T>.SetValue(const AValue: T);
begin
  FValue := AValue;
end;

{*
  [@inheritDoc]
*}
procedure TTreeNode<T>.AfterConstruction;
begin
  inherited;

  if Parent <> nil then
    Parent.AddChild(-1, Self);

  DoAncestorChanged;
end;

{*
  [@inheritDoc]
*}
procedure TTreeNode<T>.BeforeDestruction;
begin
  inherited;

  if not IsDestroying then
    PreOrderWalk(procedure(const Node: TTreeNode<T>) begin Node.Destroying end);

  if (Parent <> nil) and (not Parent.IsDestroying) then
    Parent.RemoveChild(Self);
end;

{*
  Dplace ce noeud
  @param NewParent   Nouveau parent de ce noeud (peut tre nil)
  @param Index       Index o placer ce noeud (-1 le place  la fin)
*}
procedure TTreeNode<T>.MoveTo(NewParent: TTreeNode<T>; Index: Integer = -1);
begin
  if (Parent = nil) and (NewParent = nil) then
    Exit;

  if Parent = NewParent then
  begin
    if Index <> IndexAsChild then
    begin
      Parent.FChildren.Remove(Self);
      if Index < 0 then
        Parent.FChildren.Add(Self)
      else
        Parent.FChildren.Insert(Index, Self);
    end;
  end else
  begin
    if Parent <> nil then
      Parent.RemoveChild(Self);

    FParent := NewParent;

    if Parent <> nil then
      Parent.AddChild(Index, Self);

    DoAncestorChanged;
  end;
end;

{*
  Dplace ce noeud parmi ses frres
  @param Index   Nouvel index de ce noeud (-1 le place  la fin)
*}
procedure TTreeNode<T>.MoveTo(Index: Integer);
begin
  MoveTo(Parent, Index);
end;

{*
  Cherche un noeud enfant
  @param Child   Enfant  rechercher
  @return Index de cet enfant, ou -1 si non trouv
*}
function TTreeNode<T>.IndexOf(Child: TTreeNode<T>): Integer;
begin
  Result := FChildren.IndexOf(Child);
end;

{*
  Parcours prfixe sur les noeuds de l'arbre
  @param Action   Action  effectuer pour chaque noeud
*}
procedure TTreeNode<T>.PreOrderWalk(const Action: TValueCallBack<TTreeNode<T>>);
var
  Index: Integer;
begin
  Action(Self);
  for Index := 0 to ChildCount-1 do
    Children[Index].PreOrderWalk(Action);
end;

{*
  Parcours prfixe sur les valeurs des noeuds de l'arbre
  @param Action   Action  effectuer pour chaque valeur
*}
procedure TTreeNode<T>.PreOrderWalk(const Action: TValueCallBack<T>);
begin
  PreOrderWalk(
    procedure(const Node: TTreeNode<T>)
    begin
      Action(Node.Value);
    end);
end;

{*
  Parcours postfixe sur les noeuds de l'arbre
  @param Action   Action  effectuer pour chaque noeud
*}
procedure TTreeNode<T>.PostOrderWalk(
  const Action: TValueCallBack<TTreeNode<T>>);
var
  Index: Integer;
begin
  for Index := 0 to ChildCount-1 do
    Children[Index].PostOrderWalk(Action);
  Action(Self);
end;

{*
  Parcours postfixe sur les valeurs des noeuds de l'arbre
  @param Action   Action  effectuer pour chaque valeur
*}
procedure TTreeNode<T>.PostOrderWalk(const Action: TValueCallBack<T>);
begin
  PostOrderWalk(
    procedure(const Node: TTreeNode<T>)
    begin
      Action(Node.Value);
    end);
end;

{--------------------}
{ TObjectTreeNode<T> }
{--------------------}

{*
  [@inheritDoc]
*}
destructor TObjectTreeNode<T>.Destroy;
begin
  Value.Free;
  inherited;
end;

{--------------------------}
{ TCreateObjectTreeNode<T> }
{--------------------------}

{*
  [@inheritDoc]
*}
constructor TCreateObjectTreeNode<T>.Create(AParent: TTreeNode<T>);
begin
  Create(AParent, T.Create);
end;

{*
  [@inheritDoc]
*}
constructor TCreateObjectTreeNode<T>.Create;
begin
  Create(T.Create);
end;

end.

